# -*- coding: iso-8859-1 -*-
# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.

import unittest
import itertools
from datetime import date, datetime

from logilab.common.testlib import TestCase

from rql import nodes, stmts, parse, BadRQLQuery, RQLHelper, RQLException, CoercionError
from rql.stcheck import RQLSTAnnotator

from unittest_analyze import DummySchema

schema = DummySchema()
annotator = RQLSTAnnotator(schema, {})
helper = RQLHelper(schema, None, {"eid": "uid"})


def sparse(rql):
    tree = helper.parse(rql)
    helper.compute_solutions(tree)
    return tree


class EtypeFromPyobjTC(TestCase):
    def test_bool(self):
        self.assertEqual(nodes.etype_from_pyobj(True), "Boolean")
        self.assertEqual(nodes.etype_from_pyobj(False), "Boolean")

    def test_int(self):
        self.assertEqual(nodes.etype_from_pyobj(0), "Int")

    def test_float(self):
        self.assertEqual(nodes.etype_from_pyobj(0.0), "Float")

    def test_datetime(self):
        self.assertEqual(nodes.etype_from_pyobj(datetime.now()), "Datetime")
        self.assertEqual(nodes.etype_from_pyobj(date.today()), "Date")

    def test_string(self):
        self.assertEqual(nodes.etype_from_pyobj("hop"), "String")


class TypesRestrictionNodesTest(TestCase):
    def setUp(self):
        self.parse = helper.parse
        self.simplify = helper.simplify

    def test_add_is_type_restriction(self):
        tree = self.parse("Any X WHERE X is Person, X name ILIKE 'A%'")
        select = tree.children[0]
        x = next(select.get_selected_variables())
        self.assertRaises(
            RQLException, select.add_type_restriction, x.variable, "Babar"
        )
        select.add_type_restriction(x.variable, "Person")
        self.assertEqual(tree.as_string(), 'Any X WHERE X is Person, X name ILIKE "A%"')

    def test_add_new_is_type_restriction_in(self):
        tree = self.parse("Any X WHERE X is IN(Person, Company), X name ILIKE 'A%'")
        select = tree.children[0]
        x = next(select.get_selected_variables())
        select.add_type_restriction(x.variable, "Company")
        # implementation is KISS (the IN remains)
        self.assertEqual(
            tree.as_string(), 'Any X WHERE X is IN(Company), X name ILIKE "A%"'
        )

    def test_add_new_is_type_restriction_in_nonregr(self):
        tree = self.parse("Any X WHERE X is IN(Person, Company, Student)")
        select = tree.children[0]
        x = next(select.get_selected_variables())
        select.add_type_restriction(x.variable, "Person")
        self.assertEqual(tree.as_string(), "Any X WHERE X is IN(Person)")

    def test_add_is_in_type_restriction(self):
        tree = self.parse("Any X WHERE X is IN(Person, Company), X name ILIKE 'A%'")
        select = tree.children[0]
        x = next(select.get_selected_variables())
        self.assertRaises(
            RQLException, select.add_type_restriction, x.variable, "Babar"
        )
        self.assertEqual(
            tree.as_string(),
            'Any X WHERE X is IN(Person, Company), X name ILIKE "A%"',
        )

    def test_add_is_in_type_restriction_multiple(self):
        tree = self.parse(
            "Any X WHERE X is IN(Person, Company, Student, Address), "
            " X name ILIKE 'A%'"
        )
        select = tree.children[0]
        x = next(select.get_selected_variables())
        add_restr = select.add_type_restriction
        self.assertRaises(RQLException, add_restr, x.variable, "Babar")
        self.assertRaises(RQLException, add_restr, x.variable, ["Babar"])
        self.assertRaises(RQLException, add_restr, x.variable, ["Babar", "Person"])
        add_restr(x.variable, ["Company", "Student"])
        self.assertEqual(
            tree.as_string(), 'Any X WHERE X is IN(Company, Student), X name ILIKE "A%"'
        )

    def test_add_is_type_restriction_on_is_instance_of(self):
        select = self.parse(
            "Any X WHERE X is_instance_of Person, X name ILIKE 'A%'"
        ).children[0]
        x = next(select.get_selected_variables())
        select.add_type_restriction(x.variable, "Person")
        self.assertEqual(
            select.as_string(), 'Any X WHERE X name ILIKE "A%", X is Person'
        )

    def test_add_new_is_type_restriction_in_on_is_instance_of(self):
        tree = self.parse(
            "Any X WHERE X is_instance_of IN(Person, Company), X name ILIKE 'A%'"
        )
        select = tree.children[0]
        x = next(select.get_selected_variables())
        select.add_type_restriction(x.variable, "Company")
        self.assertEqual(
            tree.as_string(), 'Any X WHERE X name ILIKE "A%", X is Company'
        )


class NodesTest(TestCase):
    def _parse(self, rql, normrql=None):
        tree = parse(rql + ";")
        tree.check_references()
        if normrql is None:
            normrql = rql
        self.assertEqual(tree.as_string(), normrql)
        # just check repr() doesn't raise an exception
        repr(tree)
        copy = tree.copy()
        self.assertEqual(copy.as_string(), normrql)
        copy.check_references()
        return tree

    def _simpleparse(self, rql):
        return self._parse(rql).children[0]

    def check_equal_but_not_same(self, tree1, tree2):
        # d1 = tree1.__dict__.copy()
        # del d1['parent']; del d1['children'] # parent and children are slots now
        # d2 = tree2.__dict__.copy()
        # del d2['parent']; del d2['children']
        self.assertNotEqual(id(tree1), id(tree2))
        self.assertTrue(tree1.is_equivalent(tree2))
        # self.assertEqual(len(tree1.children), len(tree2.children))
        # for i in range(len(tree1.children)):
        #    self.check_equal_but_not_same(tree1.children[i], tree2.children[i])

    # selection tests #########################################################

    def test_union_set_limit_1(self):
        tree = self._parse("Any X WHERE X is Person")
        select = tree.children[0]
        self.assertRaises(BadRQLQuery, tree.set_limit, 0)
        self.assertRaises(BadRQLQuery, tree.set_limit, -1)
        self.assertRaises(BadRQLQuery, tree.set_limit, "1")
        tree.save_state()
        tree.set_limit(10)
        self.assertEqual(select.limit, 10)
        self.assertEqual(tree.as_string(), "Any X LIMIT 10 WHERE X is Person")
        tree.recover()
        self.assertEqual(select.limit, None)
        self.assertEqual(tree.as_string(), "Any X WHERE X is Person")

    def test_union_set_limit_2(self):
        # not undoable set_limit since a new root has to be introduced
        tree = self._parse("(Any X WHERE X is Person) UNION (Any X WHERE X is Company)")
        tree.save_state()
        tree.set_limit(10)
        self.assertEqual(
            tree.as_string(),
            "Any A LIMIT 10 WITH A BEING "
            "((Any X WHERE X is Person) UNION (Any X WHERE X is Company))",
        )
        select = tree.children[0]
        self.assertEqual(select.limit, 10)
        tree.recover()
        self.assertEqual(
            tree.as_string(),
            "(Any X WHERE X is Person) UNION (Any X WHERE X is Company)",
        )

    def test_union_set_offset_1(self):
        tree = self._parse("Any X WHERE X is Person")
        select = tree.children[0]
        self.assertRaises(BadRQLQuery, tree.set_offset, -1)
        self.assertRaises(BadRQLQuery, tree.set_offset, "1")
        tree.save_state()
        tree.set_offset(10)
        self.assertEqual(select.offset, 10)
        tree.recover()
        self.assertEqual(select.offset, 0)
        self.assertEqual(tree.as_string(), "Any X WHERE X is Person")

    def test_union_set_offset_2(self):
        # not undoable set_offset since a new root has to be introduced
        tree = self._parse("(Any X WHERE X is Person) UNION (Any X WHERE X is Company)")
        tree.save_state()
        tree.set_offset(10)
        select = tree.children[0]
        self.assertEqual(
            tree.as_string(),
            "Any A OFFSET 10 WITH A BEING "
            "((Any X WHERE X is Person) UNION (Any X WHERE X is Company))",
        )
        self.assertEqual(select.offset, 10)
        tree.recover()
        self.assertEqual(
            tree.as_string(),
            "(Any X WHERE X is Person) UNION (Any X WHERE X is Company)",
        )

    def test_union_undo_add_rel(self):
        tree = self._parse(
            "Any A WITH A BEING ((Any X WHERE X is Person) UNION (Any X WHERE X is Company))"
        )
        tree.save_state()
        select = tree.children[0]
        var = select.make_variable()
        mainvar = select.selection[0].variable
        select.add_relation(mainvar, "name", var)
        self.assertEqual(
            tree.as_string(),
            "Any A WHERE A name B WITH A BEING "
            "((Any X WHERE X is Person) UNION (Any X WHERE X is Company))",
        )
        tree.recover()
        self.assertEqual(
            tree.as_string(),
            "Any A WITH A BEING ((Any X WHERE X is Person) UNION (Any X WHERE X is Company))",
        )

    def test_select_set_limit(self):
        tree = self._simpleparse("Any X WHERE X is Person")
        self.assertEqual(tree.limit, None)
        self.assertRaises(BadRQLQuery, tree.set_limit, 0)
        self.assertRaises(BadRQLQuery, tree.set_limit, -1)
        self.assertRaises(BadRQLQuery, tree.set_limit, "1")
        tree.save_state()
        tree.set_limit(10)
        self.assertEqual(tree.limit, 10)
        tree.recover()
        self.assertEqual(tree.limit, None)

    def test_select_set_offset(self):
        tree = self._simpleparse("Any X WHERE X is Person")
        self.assertRaises(BadRQLQuery, tree.set_offset, -1)
        self.assertRaises(BadRQLQuery, tree.set_offset, "1")
        self.assertEqual(tree.offset, 0)
        tree.save_state()
        tree.set_offset(0)
        self.assertEqual(tree.offset, 0)
        tree.set_offset(10)
        self.assertEqual(tree.offset, 10)
        tree.recover()
        self.assertEqual(tree.offset, 0)

    def test_select_add_sort_var(self):
        tree = self._parse("Any X")
        tree.save_state()
        select = tree.children[0]
        select.add_sort_var(select.get_variable("X"))
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X ORDERBY X")
        tree.recover()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X")

    def test_select_remove_sort_terms(self):
        tree = self._parse("Any X,Y ORDERBY X,Y")
        tree.save_state()
        select = tree.children[0]
        select.remove_sort_terms()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X,Y")
        tree.recover()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X,Y ORDERBY X,Y")

    def test_select_undefine_variable(self):
        tree = sparse("Any X,Y ORDERBY X,Y WHERE X work_for Y")
        tree.save_state()
        select = tree.children[0]
        select.undefine_variable(select.defined_vars["Y"])
        self.assertCountEqual(select.solutions, [{"X": "Person"}, {"X": "Student"}])
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X ORDERBY X")
        tree.recover()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X,Y ORDERBY X,Y WHERE X work_for Y")
        self.assertCountEqual(
            select.solutions,
            [{"X": "Person", "Y": "Company"}, {"X": "Student", "Y": "Company"}],
        )

    def test_select_set_distinct(self):
        tree = self._parse("DISTINCT Any X")
        tree.save_state()
        select = tree.children[0]
        self.assertEqual(select.distinct, True)
        tree.save_state()
        select.set_distinct(True)
        self.assertEqual(select.distinct, True)
        tree.recover()
        self.assertEqual(select.distinct, True)
        select.set_distinct(False)
        self.assertEqual(select.distinct, False)
        tree.recover()
        self.assertEqual(select.distinct, True)

    def test_select_add_group_var(self):
        tree = self._parse("Any X")
        tree.save_state()
        select = tree.children[0]
        select.add_group_var(select.get_variable("X"))
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X GROUPBY X")
        tree.recover()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X")

    def test_select_remove_group_term(self):
        tree = self._parse("Any X GROUPBY X")
        tree.save_state()
        select = tree.children[0]
        select.remove_group_term(select.groupby[0])
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X")
        tree.recover()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X GROUPBY X")

    def test_select_remove_groups(self):
        tree = self._parse("Any X,Y GROUPBY X,Y")
        tree.save_state()
        select = tree.children[0]
        select.remove_groups()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X,Y")
        tree.recover()
        tree.check_references()
        self.assertEqual(tree.as_string(), "Any X,Y GROUPBY X,Y")

    def test_recover_add_type_restriction_is_in(self):
        tree = self._parse('Any X WHERE X is IN(Person, Company), X name ILIKE "A%"')
        annotator.annotate(tree)  # needed to get typerel index
        tree.save_state()
        select = tree.children[0]
        x = next(select.get_selected_variables())
        select.add_type_restriction(x.variable, "Company")
        self.assertEqual(
            tree.as_string(),
            'Any X WHERE X is IN(Company), X name ILIKE "A%"',
        )
        tree.recover()
        tree.check_references()
        self.assertEqual(
            tree.as_string(),
            'Any X WHERE X is IN(Person, Company), X name ILIKE "A%"',
        )

    def test_recover_add_type_restriction_is_instance_of(self):
        tree = self._parse(
            'Any X WHERE X is_instance_of IN(Person, Company), X name ILIKE "A%"'
        )
        annotator.annotate(tree)  # needed to get typerel index
        tree.save_state()
        select = tree.children[0]
        x = next(select.get_selected_variables())
        select.add_type_restriction(x.variable, "Company")
        self.assertEqual(
            tree.as_string(), 'Any X WHERE X name ILIKE "A%", X is Company'
        )
        tree.recover()
        tree.check_references()
        self.assertEqual(
            tree.as_string(),
            'Any X WHERE X is_instance_of IN(Person, Company), X name ILIKE "A%"',
        )

    def test_select_base_1(self):
        tree = self._parse("Any X WHERE X is Person")
        self.assertIsInstance(tree, stmts.Union)
        select = tree.children[0]
        self.assertIsNone(select.limit)
        self.assertEqual(select.offset, 0)
        self.assertIsInstance(select, stmts.Select)
        self.assertEqual(select.distinct, False)
        self.assertEqual(len(select.children), 2)
        self.assertIs(select.children[0], select.selection[0])
        self.assertIs(select.children[1], select.where)
        self.assertIsInstance(select.where, nodes.Relation)

    def test_select_base_2(self):
        tree = self._simpleparse("Any X WHERE X is Person")
        self.assertEqual(len(tree.children), 2)
        self.assertEqual(tree.distinct, False)

    def test_select_distinct(self):
        tree = self._simpleparse("DISTINCT Any X WHERE X is Person")
        self.assertEqual(len(tree.children), 2)
        self.assertEqual(tree.distinct, True)

    def test_select_null(self):
        tree = self._simpleparse("Any X WHERE X name NULL")
        constant = tree.where.children[1].children[0]
        self.assertEqual(constant.type, None)
        self.assertEqual(constant.value, None)

    def test_select_true(self):
        tree = self._simpleparse("Any X WHERE X name TRUE")
        constant = tree.where.children[1].children[0]
        self.assertEqual(constant.type, "Boolean")
        self.assertEqual(constant.value, True)

    def test_select_false(self):
        tree = self._simpleparse("Any X WHERE X name FALSE")
        constant = tree.where.children[1].children[0]
        self.assertEqual(constant.type, "Boolean")
        self.assertEqual(constant.value, False)

    def test_select_date(self):
        tree = self._simpleparse("Any X WHERE X born TODAY")
        constant = tree.where.children[1].children[0]
        self.assertEqual(constant.type, "Date")
        self.assertEqual(constant.value, "TODAY")

    def test_select_int(self):
        tree = self._simpleparse("Any X WHERE X name 1")
        constant = tree.where.children[1].children[0]
        self.assertEqual(constant.type, "Int")
        self.assertEqual(constant.value, 1)

    def test_select_float(self):
        tree = self._simpleparse("Any X WHERE X name 1.0")
        constant = tree.where.children[1].children[0]
        self.assertEqual(constant.type, "Float")
        self.assertEqual(constant.value, 1.0)

    def test_select_group(self):
        tree = self._simpleparse("Any X GROUPBY N WHERE X is Person, X name N")
        self.assertEqual(tree.distinct, False)
        self.assertEqual(len(tree.children), 3)
        self.assertIsInstance(tree.where, nodes.And)
        self.assertIsInstance(tree.groupby[0], nodes.VariableRef)
        self.assertEqual(tree.groupby[0].name, "N")

    def test_select_ord_default(self):
        tree = self._parse("Any X ORDERBY N WHERE X is Person, X name N")
        self.assertEqual(tree.children[0].orderby[0].asc, 1)

    def test_select_ord_desc(self):
        tree = self._parse("Any X ORDERBY N DESC WHERE X is Person, X name N")
        select = tree.children[0]
        self.assertEqual(len(select.children), 3)
        self.assertIsInstance(select.where, nodes.And)
        sort = select.orderby
        self.assertIsInstance(sort[0], nodes.SortTerm)
        self.assertEqual(sort[0].term.name, "N")
        self.assertEqual(sort[0].asc, 0)
        self.assertEqual(sort[0].nulls_sort, 0)
        self.assertEqual(select.distinct, False)

    def test_select_ord_nullsfirst(self):
        tree = self._parse(
            "Any X ORDERBY N DESC NULLSFIRST WHERE X is Person, X name N"
        )
        select = tree.children[0]
        self.assertEqual(len(select.children), 3)
        self.assertIsInstance(select.where, nodes.And)
        sort = select.orderby
        self.assertIsInstance(sort[0], nodes.SortTerm)
        self.assertEqual(sort[0].term.name, "N")
        self.assertEqual(sort[0].asc, 0)
        self.assertEqual(sort[0].nulls_sort, 1)
        self.assertEqual(select.distinct, False)

    def test_select_ord_nullslast(self):
        tree = self._parse("Any X ORDERBY N DESC NULLSLAST WHERE X is Person, X name N")
        select = tree.children[0]
        self.assertEqual(len(select.children), 3)
        self.assertIsInstance(select.where, nodes.And)
        sort = select.orderby
        self.assertIsInstance(sort[0], nodes.SortTerm)
        self.assertEqual(sort[0].term.name, "N")
        self.assertEqual(sort[0].asc, 0)
        self.assertEqual(sort[0].nulls_sort, 2)
        self.assertEqual(select.distinct, False)

    def test_select_ord_multi_nullslast(self):
        tree = self._parse(
            "Any X ORDERBY N DESC NULLSLAST,X NULLSFIRST WHERE X is Person, X name N"
        )
        select = tree.children[0]
        self.assertEqual(len(select.children), 4)
        self.assertIsInstance(select.where, nodes.And)
        sort = select.orderby
        self.assertIsInstance(sort[0], nodes.SortTerm)
        self.assertEqual(sort[0].term.name, "N")
        self.assertEqual(sort[0].asc, 0)
        self.assertEqual(sort[0].nulls_sort, 2)
        self.assertIsInstance(sort[1], nodes.SortTerm)
        self.assertEqual(sort[1].term.name, "X")
        self.assertEqual(sort[1].asc, 1)
        self.assertEqual(sort[1].nulls_sort, 1)
        self.assertEqual(select.distinct, False)

    def test_select_group_ord_asc(self):
        tree = self._parse(
            "Any X GROUPBY N ORDERBY N ASC WHERE X is Person, X name N",
            "Any X GROUPBY N ORDERBY N WHERE X is Person, X name N",
        )
        select = tree.children[0]
        self.assertEqual(len(select.children), 4)
        group = select.groupby
        self.assertIsInstance(group[0], nodes.VariableRef)
        self.assertEqual(group[0].name, "N")

    def test_select_limit_offset(self):
        tree = self._parse("Any X LIMIT 10 OFFSET 10 WHERE X name 1.0")
        select = tree.children[0]
        self.assertEqual(select.limit, 10)
        self.assertEqual(select.offset, 10)

    def test_exists(self):
        self._simpleparse("Any X,N WHERE X is Person, X name N, EXISTS(X work_for Y)")

    def test_copy(self):
        tree = self._parse(
            "Any X,LOWER(Y) GROUPBY N ORDERBY N WHERE X is Person, X name N, X date >= TODAY"
        )
        select = stmts.Select()
        restriction = tree.children[0].where
        self.check_equal_but_not_same(restriction, restriction.copy(select))
        group = tree.children[0].groupby[0]
        self.check_equal_but_not_same(group, group.copy(select))
        sort = tree.children[0].orderby[0]
        self.check_equal_but_not_same(sort, sort.copy(select))

    def test_selected_index(self):
        tree = self._simpleparse("Any X ORDERBY N DESC WHERE X is Person, X name N")
        annotator.annotate(tree)
        self.assertEqual(tree.defined_vars["X"].selected_index(), 0)
        self.assertEqual(tree.defined_vars["N"].selected_index(), None)

    def test_get_variable_indices_1(self):
        dummy = self._parse("Any A,B,C")
        dummy.children[0].solutions = [
            {"A": "String", "B": "EUser", "C": "EGroup"},
            {"A": "String", "B": "Personne", "C": "EGroup"},
            {"A": "String", "B": "EUser", "C": "Societe"},
        ]
        self.assertEqual(dummy.get_variable_indices(), set([1, 2]))

    def test_get_variable_indices_2(self):
        dummy = self._parse("Any A,B WHERE B relation C")
        dummy.children[0].solutions = [
            {"A": "String", "B": "EUser", "C": "EGroup"},
            {"A": "String", "B": "Personne", "C": "EGroup"},
            {"A": "String", "B": "EUser", "C": "Societe"},
        ]
        self.assertEqual(dummy.get_variable_indices(), set([1]))

    def test_get_variable_indices_3(self):
        dummy = self._parse("(Any X WHERE X is EGroup) UNION (Any C WHERE C is EUser)")
        dummy.children[0].solutions = [{"X": "EGroup"}]
        dummy.children[1].solutions = [{"C": "EUser"}]
        self.assertEqual(dummy.get_variable_indices(), set([0]))

    def test_get_variable_indices_4(self):
        dummy = self._parse(
            "(Any X,XN WHERE X is EGroup, X name XN) UNION (Any C,CL WHERE C is EUser, C login CL)"
        )
        dummy.children[0].solutions = [{"X": "EGroup", "XN": "String"}]
        dummy.children[1].solutions = [{"C": "EUser", "CL": "String"}]
        self.assertEqual(dummy.get_variable_indices(), set([0]))

    # insertion tests #########################################################

    def test_insert_base_1(self):
        tree = self._parse("INSERT Person X")
        self.assertIsInstance(tree, stmts.Insert)
        self.assertEqual(len(tree.children), 1)
        # test specific attributes
        self.assertEqual(len(tree.main_relations), 0)
        self.assertEqual(len(tree.main_variables), 1)
        self.assertEqual(tree.main_variables[0][0], "Person")
        self.assertIsInstance(tree.main_variables[0][1], nodes.VariableRef)
        self.assertEqual(tree.main_variables[0][1].name, "X")

    def test_insert_base_2(self):
        tree = self._parse('INSERT Person X : X name "bidule"')
        # test specific attributes
        self.assertEqual(len(tree.main_relations), 1)
        self.assertIsInstance(tree.main_relations[0], nodes.Relation)
        self.assertEqual(len(tree.main_variables), 1)
        self.assertEqual(tree.main_variables[0][0], "Person")
        self.assertIsInstance(tree.main_variables[0][1], nodes.VariableRef)
        self.assertEqual(tree.main_variables[0][1].name, "X")

    def test_insert_multi(self):
        tree = self._parse(
            'INSERT Person X, Person Y : X name "bidule", Y name "chouette", X friend Y'
        )
        # test specific attributes
        self.assertEqual(len(tree.main_relations), 3)
        for relation in tree.main_relations:
            self.assertIsInstance(relation, nodes.Relation)
        self.assertEqual(len(tree.main_variables), 2)
        self.assertEqual(tree.main_variables[0][0], "Person")
        self.assertIsInstance(tree.main_variables[0][1], nodes.VariableRef)
        self.assertEqual(tree.main_variables[0][1].name, "X")
        self.assertEqual(tree.main_variables[1][0], "Person")
        self.assertIsInstance(tree.main_variables[0][1], nodes.VariableRef)
        self.assertEqual(tree.main_variables[1][1].name, "Y")

    def test_insert_where(self):
        tree = self._parse(
            'INSERT Person X : X name "bidule", X friend Y WHERE Y name "chouette"'
        )
        self.assertEqual(len(tree.children), 4)
        self.assertIsInstance(tree.where, nodes.Relation)
        # test specific attributes
        self.assertEqual(len(tree.main_relations), 2)
        for relation in tree.main_relations:
            self.assertIsInstance(relation, nodes.Relation)
        self.assertEqual(len(tree.main_variables), 1)
        self.assertEqual(tree.main_variables[0][0], "Person")
        self.assertIsInstance(tree.main_variables[0][1], nodes.VariableRef)
        self.assertEqual(tree.main_variables[0][1].name, "X")

    # update tests ############################################################

    def test_update_1(self):
        tree = self._parse('SET X name "toto" WHERE X is Person, X name "bidule"')
        self.assertIsInstance(tree, stmts.Set)
        self.assertEqual(len(tree.children), 2)
        self.assertIsInstance(tree.where, nodes.And)

    # deletion tests #########################################################

    def test_delete_1(self):
        tree = self._parse('DELETE Person X WHERE X name "toto"')
        self.assertIsInstance(tree, stmts.Delete)
        self.assertEqual(len(tree.children), 2)
        self.assertIsInstance(tree.where, nodes.Relation)

    def test_delete_2(self):
        tree = self._parse('DELETE X friend Y WHERE X name "toto"')
        self.assertIsInstance(tree, stmts.Delete)
        self.assertEqual(len(tree.children), 2)
        self.assertIsInstance(tree.where, nodes.Relation)

    # as_string tests ####################################################

    def test_as_string(self):
        tree = parse("SET X know Y WHERE X friend Y;")
        self.assertEqual(tree.as_string(), "SET X know Y WHERE X friend Y")

        tree = parse("Person X")
        self.assertEqual(tree.as_string(), "Any X WHERE X is Person")

        tree = parse("Any X WHERE X has_text 'héhé'")
        self.assertEqual(tree.as_string(), 'Any X WHERE X has_text "héhé"')
        tree = parse("Any X WHERE X has_text %(text)s")
        self.assertEqual(
            tree.as_string({"text": "héhé"}), 'Any X WHERE X has_text "héhé"'
        )
        tree = parse("Any X WHERE X has_text %(text)s")
        self.assertEqual(
            tree.as_string({"text": 'hé"hé'}), 'Any X WHERE X has_text "hé\\"hé"'
        )
        tree = parse("Any X WHERE X has_text %(text)s")
        self.assertEqual(
            tree.as_string({"text": "hé\"'hé"}), 'Any X WHERE X has_text "hé\\"\'hé"'
        )

    def test_get_type_is_in(self):
        tree = sparse("Any X WHERE X is IN (Person, Company)")
        select = tree.children[0]
        self.assertEqual(select.defined_vars["X"].get_type(), "Any")

    # sub-queries tests #######################################################

    def test_subq_colalias_compat(self):
        tree = sparse(
            "Any X ORDERBY N WHERE X creation_date <NOW WITH X,N BEING ("
            "(Any X,N WHERE X firstname N) UNION (Any X,N WHERE X name N, X is Company))"
        )
        select = tree.children[0]
        select.save_state()
        select.remove_sort_term(select.orderby[0])
        select.recover()
        X = select.get_variable("X")
        N = select.get_variable("N")
        self.assertEqual(len(X.references()), 3)
        self.assertEqual(len(N.references()), 2)
        tree.schema = schema
        # annotator.annotate(tree)
        # XXX how to choose
        self.assertEqual(X.get_type(), "Company")
        self.assertEqual(X.get_type({"X": "Person"}), "Person")
        # self.assertEqual(N.get_type(), 'String')
        self.assertEqual(
            X.get_description(0, lambda x, **k: x), "Company, Person, Student"
        )
        self.assertEqual(N.get_description(0, lambda x, **k: x), "firstname, name")
        self.assertEqual(X.selected_index(), 0)
        self.assertEqual(N.selected_index(), None)
        self.assertEqual(X.main_relation(), None)

    # non regression tests ####################################################

    def test_get_description_and_get_type(self):
        tree = sparse(
            "Any N,COUNT(X),NOW-D GROUPBY N WHERE X name N, X creation_date D;"
        )
        tree.schema = schema
        self.assertEqual(
            tree.get_description(),
            [["name", "COUNT(Company, Person, Student)", "creation_date"]],
        )
        select = tree.children[0]
        self.assertEqual(select.selection[0].get_type(), "Any")
        self.assertEqual(select.selection[1].get_type(), "Int")
        self.assertEqual(
            select.defined_vars["D"].get_type({"D": "Datetime"}), "Datetime"
        )
        self.assertEqual(select.selection[2].get_type({"D": "Datetime"}), "Interval")

    def test_date_arithmetic(self):
        minus_expr = sparse("Any D1-D2;").children[0].selection[0]
        plus_expr = sparse("Any D1+D2;").children[0].selection[0]
        for d1t, d2t in itertools.combinations_with_replacement(
            ["Date", "Datetime", "TZDatetime"], 2
        ):
            self.assertEqual(minus_expr.get_type({"D1": d1t, "D2": d2t}), "Interval")
            with self.assertRaises(CoercionError):
                plus_expr.get_type({"D1": d1t, "D2": d2t})
        for d1t in ("Date", "Datetime", "TZDatetime"):
            expected_type = "Datetime" if d1t == "Date" else d1t
            self.assertEqual(
                minus_expr.get_type({"D1": d1t, "D2": "Interval"}), expected_type
            )
            self.assertEqual(
                plus_expr.get_type({"D1": d1t, "D2": "Interval"}), expected_type
            )

    def test_get_description_simplified(self):
        tree = sparse("Any X,R,D WHERE X eid 2, X work_for R, R creation_date D")
        self.assertEqual(
            tree.get_description(), [["work_for", "work_for_object", "creation_date"]]
        )
        helper.simplify(tree)
        # Any since const.uid_type is used while solutions have not been computed
        self.assertEqual(
            tree.get_description(), [["Any", "work_for_object", "creation_date"]]
        )

    def test_repr_encoding(self):
        tree = parse('Any N where NOT N has_text "bidüle"')
        repr(tree)

    def test_get_description_mainvar_objrel(self):
        tree = sparse("Any X,R,D,Y WHERE X work_for R, R creation_date D, Y owned_by X")
        self.assertEqual(
            tree.get_description(0),
            [["Person", "work_for", "creation_date", "owned_by_object"]],
        )

    def test_get_description_mainvar_symrel(self):
        tree = sparse("Any X,R,D,Y WHERE X work_for R, R creation_date D, Y connait X")
        self.assertEqual(
            tree.get_description(0),
            [["Person, Student", "work_for", "creation_date", "connait"]],
        )

    def test_get_description_cast(self):
        tree = sparse("Any CAST(String, Y) WHERE X creation_date Y")
        select = tree.children[0]
        self.assertEqual(select.selection[0].get_type(), "String")
        self.assertEqual(tree.get_description(0), [["String"]])

    def test_replace_node_select(self):
        tree = sparse('Any X, MAX(X) WHERE X name "toto tata"')
        select = tree.children[0]
        funcnode = sparse("Any MAX(X)").children[0].selection[0]
        select.replace(funcnode, nodes.Constant(1.0, "Float"))

    def test_replace_node_orderby(self):
        tree = sparse('Any X ORDERBY MAX(X) WHERE X name "toto tata"')
        select = tree.children[0]
        funcnode = sparse("Any X ORDERBY MAX(X)").children[0].orderby[0]
        select.replace(funcnode, nodes.Constant(1.0, "Float"))

    def test_replace_node_groupby(self):
        tree = sparse('Any X GROUPBY X, MAX(X) WHERE X name "toto tata"')
        select = tree.children[0]
        funcnode = sparse("Any MAX(X)").children[0].selection[0]
        select.replace(funcnode, nodes.Constant(1.0, "Float"))

    def test_replace_node_having(self):
        tree = sparse('Any X WHERE X name "toto tata" HAVING MAX(X) > 1')
        select = tree.children[0]
        funcnode = sparse("Any X HAVING MAX(X) > 1").children[0].having[0]
        select.replace(funcnode, nodes.Constant(1.0, "Float"))

    def test_undo_node_having(self):
        qs = "Any X WHERE X name N"
        tree = sparse(qs)
        select = tree.children[0]
        select.save_state()
        namevar = select.where.relation().children[-1].children[-1].variable
        comp = nodes.Comparison(">")
        maxf = nodes.Function("MAX")
        maxf.append(nodes.VariableRef(namevar))
        comp.append(maxf)
        comp.append(nodes.Constant(1, "Int"))
        select.set_having([comp])
        select.recover()
        self.assertEqual(select.as_string(), qs)

    def test_undo_node_having_in_mutable(self):
        sparse("DELETE Person X WHERE X name N HAVING MAX(X) > 1")
        sparse('SET X name "foo" WHERE X is Person HAVING MAX(X) > 1')


class GetNodesFunctionTest(TestCase):
    def test_known_values_1(self):
        tree = parse('Any X where X name "turlututu"').children[0]
        constants = tree.get_nodes(nodes.Constant)
        self.assertEqual(len(constants), 1)
        self.assertEqual(isinstance(constants[0], nodes.Constant), 1)
        self.assertEqual(constants[0].value, "turlututu")

    def test_known_values_2(self):
        tree = parse(
            'Any X where X name "turlututu", Y know X, Y name "chapo pointu"'
        ).children[0]
        varrefs = tree.get_nodes(nodes.VariableRef)
        self.assertEqual(len(varrefs), 5)
        for varref in varrefs:
            self.assertIsInstance(varref, nodes.VariableRef)
        self.assertEqual(sorted(x.name for x in varrefs), ["X", "X", "X", "Y", "Y"])

    def test_iknown_values_1(self):
        tree = parse('Any X where X name "turlututu"').children[0]
        constants = list(tree.iget_nodes(nodes.Constant))
        self.assertEqual(len(constants), 1)
        self.assertEqual(isinstance(constants[0], nodes.Constant), 1)
        self.assertEqual(constants[0].value, "turlututu")

    def test_iknown_values_2(self):
        tree = parse(
            'Any X where X name "turlututu", Y know X, Y name "chapo pointu"'
        ).children[0]
        varrefs = list(tree.iget_nodes(nodes.VariableRef))
        self.assertEqual(len(varrefs), 5)
        for varref in varrefs:
            self.assertIsInstance(varref, nodes.VariableRef)
        self.assertEqual(sorted(x.name for x in varrefs), ["X", "X", "X", "Y", "Y"])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# pylint: disable-msg=W0404,W0622,W0704,W0613,E0611,C0103
# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.
"""Generic Setup script, takes package info from __pkginfo__.py file.
"""

from setuptools import setup, find_packages
from setuptools.command import build_ext
from io import open
import os
import os.path as osp
import sys

here = osp.abspath(osp.dirname(__file__))

pkginfo = {}
with open(osp.join(here, "__pkginfo__.py")) as f:
    exec(f.read(), pkginfo)

# Get the long description from the relevant file
with open(osp.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

if os.environ.get("RQL_FORCE_GECODE"):
    MyBuildExt = build_ext.build_ext
else:

    class MyBuildExt(build_ext.build_ext):
        """Extend build_ext command to pass through compilation error.
        In fact, if gecode extension fail, rql will use logilab.constraint
        """

        def run(self):
            try:
                build_ext.build_ext.run(self)
            except Exception:
                import traceback

                traceback.print_exc()
                sys.stderr.write(
                    "================================\n"
                    "The compilation of the gecode C extension failed. "
                    "rql will use logilab.constraint which is a pure "
                    "python implementation. "
                    "Please note that the C extension run faster. "
                    "So, install a compiler then install rql again with"
                    ' the "force" option for better performance.\n'
                    "================================\n"
                )


setup(
    name=pkginfo.get("distname", pkginfo["modname"]),
    version=pkginfo["version"],
    description=pkginfo["description"],
    long_description=long_description,
    url=pkginfo["web"],
    author=pkginfo["author"],
    author_email=pkginfo["author_email"],
    license=pkginfo["license"],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=pkginfo.get("classifiers", []),
    packages=find_packages(exclude=["contrib", "docs", "test*"]),
    package_data={"rql": ["py.typed"]},
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=pkginfo["install_requires"],
    ext_modules=pkginfo.get("ext_modules"),
    cmdclass={"build_ext": MyBuildExt},
)



# -*- coding: iso-8859-1 -*-
# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.

import unittest

from yapps.runtime import print_error, SyntaxError

from rql.parser import Hercule, HerculeScanner
from rql import BadRQLQuery, RQLSyntaxError, nodes
from rql import parse


BAD_SYNTAX_QUERIES = (
    "ANY X WHERE X name Nulll;",
    'ANY X WHERE (X name NULL or X name "chouette";',
    'INSERT Person X : X name "bidule" or X name "chouette";',
    'Any X WHERE "YO" related "UPI";',
    # FIXME: incorrect because X/Y are not bound, not a syntax error
    "Personne P WHERE OFFSET 200;",
    (
        'Any X GROUPBY X ORDERBY X WHERE X nom "toto" UNION Any X GROUPBY X '
        'ORDERBY X WHERE X firstname "toto";'
    ),
    (
        '(Any X GROUPBY X WHERE X nom "toto") UNION '
        '(Any X GROUPBY X WHERE X firstname "toto") ORDERBY X;'
    ),
    "Any X, X/Y FROM (Any SUM(X) WHERE X is Person) WHERE X is Person;",  # missing AS for subquery
    "Any X, X/Y FROM (Any X WHERE X is) WHERE X is Person;",  # missing AS for subquery
)

BAD_QUERIES = (
    "Person Marcou;",
    'INSERT Any X : X name "bidule";',
    "DELETE Any X;",
    "Person Marcou",
    'INSERT Person X : Y name "bidule" WHERE X work_for Y;',
    "Any X LIMIT -1;",
    "Any X OFFSET -1;",
    "Any X ORDERBY Y;",
    "Any N,COUNT(X) GROUPBY N "
    " HAVING COUNT(X)>1 "
    " WITH X,N BEING (Any X, N WHERE X name N, X is State UNION "
    "                 Any X, N WHERE X name N, X is Transition);",
    "Machin 12",
)

# FIXME: this shoud be generated from the spec file
SPEC_QUERIES = (
    "Any X WHERE X eid 53;",
    "Any X WHERE X eid -53;",
    (
        "Document X WHERE X occurence_of F, F class C, C name 'Bande dessinée', "
        "X owned_by U, U login 'syt', X available true;"
    ),
    (
        "Document X WHERE X occurence_of F, F class C, C name 'Bande dessinée', "
        "X owned_by U, U login 'syt', X available true;"
    ),
    (
        "Personne P WHERE P travaille_pour S, S nom 'Eurocopter', P interesse_par T, "
        "T nom 'formation';"
    ),
    "Note N WHERE N ecrit_le D, D day > (today -10), N ecrit_par P, P nom 'jphc' or P nom 'ocy';",
    "Personne P WHERE (P interesse_par T, T nom 'formation') or (P ville 'Paris');",
    "Any X ORDERBY S DESC WHERE X is Person, X firstname 'Anne', X surname S;",
    # limit / offset
    "Personne P LIMIT 100 WHERE P nom N;",
    "Personne P LIMIT 100 OFFSET 200 WHERE P nom N;",
    "Any X OFFSET 6 WHERE X is Person;",
    # optional relation support (left|right outer join)
    "Any X,Y,A WHERE X? concerns Y, Y title A;",
    "Any X,Y,A WHERE X concerns Y?, Y title A;",
    "Any X,Y,A WHERE X? concerns Y?, Y title A;",
    # data modification query
    "INSERT Personne X: X nom 'bidule';",
    "INSERT Personne X, Personne Y: X nom 'bidule', Y nom 'chouette', X ami Y;",
    "INSERT Person X: X nom 'bidule', X ami Y WHERE Y nom 'chouette';",
    "SET X nom 'toto', X prenom 'original' WHERE X is Person, X nom 'bidule';",
    "SET X know Y WHERE X ami Y;",
    "SET X value -Y WHERE X value Y;",
    "DELETE Person X WHERE X nom 'toto';",
    "DELETE X ami Y WHERE X is Person, X nom 'toto';",
    # some additional cases
    'INSERT Person X : X name "bidule", Y workfor X WHERE Y name "logilab";',
    (
        "DISTINCT Any X,A,B,C,D ORDERBY A ASC WHERE P eid 41, X concerns P, "
        "P is Project, X is Story,X title A,X state B,X priority C,X cost D;"
    ),
    'Any X WHERE X has_text "2.12.0";',
    "Any X,A,B,C,D ORDERBY A ASC WHERE X concerns 41,X title A,X state B,X priority C,X cost D;",
    "Any X, COUNT(B) GROUPBY X ORDERBY 1 where B concerns X;",
    "Any X ORDERBY RANDOM();",
    "Any X ORDERBY F(1, 2);",
    "Any X, COUNT(B) GROUPBY X ORDERBY 1 WHERE B concerns X HAVING COUNT(B) > 2;",
    "Any X, MAX(COUNT(B)) GROUPBY X WHERE B concerns X;",  # syntaxically correct
    "Any X WHERE X eid > 12;",
    "DELETE Any X WHERE X eid > 12;",
    "(Any X WHERE X eid > 12) UNION (Any X WHERE X eid < 23);",
    '(Any X WHERE X nom "toto") UNION (Any X WHERE X firstname "toto");',
    (
        '(Any X GROUPBY X WHERE X nom "toto") UNION '
        '(Any X GROUPBY X ORDERBY X WHERE X firstname "toto");'
    ),
    "Any X, X/Y WHERE X is Person WITH Y BEING (Any SUM(X) WHERE X is Person);",
    "Any Y, COUNT(X) GROUPBY Y WHERE X bla Y WITH Y BEING ((Person X) UNION (Document X));",
    "Any T2, COUNT(T1)"
    " GROUPBY T1"
    " ORDERBY 2 DESC, T2;"
    " WITH T1,T2 BEING ("
    "      (Any X,N WHERE X name N, X transition_of E, E name %(name)s)"
    "       UNION "
    "      (Any X,N WHERE X name N, X state_of E, E name %(name)s));",
    "Any T2" " GROUPBY T2" " WHERE T1 relation T2" " HAVING COUNT(T1) IN (1,2);",
    "Any T2"
    " GROUPBY T2"
    " WHERE T1 relation T2"
    " HAVING COUNT(T1) IN (1,2) OR COUNT(T1) IN (3,4);",
    "Any T2"
    " GROUPBY T2"
    " WHERE T1 relation T2"
    " HAVING 1 < COUNT(T1) OR COUNT(T1) IN (3,4);",
    "Any T2"
    " GROUPBY T2"
    " WHERE T1 relation T2"
    " HAVING (COUNT(T1) IN (1,2)) OR (COUNT(T1) IN (3,4));",
    "Any T2"
    " GROUPBY T2"
    " WHERE T1 relation T2"
    " HAVING (1 < COUNT(T1) OR COUNT(T1) IN (3,4));",
    "Any T2" " GROUPBY T2" " WHERE T1 relation T2" " HAVING 1+2 < COUNT(T1);",
    "Any X,Y,A ORDERBY Y "
    "WHERE A done_for Y, X split_into Y, A diem D "
    'HAVING MIN(D) < "2010-07-01", MAX(D) >= "2010-07-01";',
    "Any YEAR(XD),COUNT(X) GROUPBY YEAR(XD) ORDERBY YEAR(XD) WHERE X date XD;",
    "Any YEAR(XD),COUNT(X) GROUPBY 1 ORDERBY 1 WHERE X date XD;",
    "Any -1.0;",
    "Any U,G WHERE U login UL, G name GL, G is CWGroup HAVING UPPER(UL)=UPPER(GL)?;",
    "Any U,G WHERE U login UL, G name GL, G is CWGroup HAVING UPPER(UL)?=UPPER(GL);",
)


class ParserHercule(unittest.TestCase):
    _syntaxerr = SyntaxError

    def parse(self, string, print_errors=False):
        try:
            parser = Hercule(HerculeScanner(string))
            return parser.goal()
        except SyntaxError as ex:
            if print_errors:
                # try to get error message from yapps
                print_error(ex, parser._scanner)
                print
            raise
        except Exception as ex:
            if print_errors:
                print(string, ex)
            raise

    def test_unicode_constant(self):
        tree = self.parse("Any X WHERE X name 'Ångström';")
        base = tree.children[0].where
        comparison = base.children[1]
        self.assertIsInstance(comparison, nodes.Comparison)
        rhs = comparison.children[0]
        self.assertIsInstance(rhs.value, str)

    def test_precedence_1(self):
        tree = self.parse(
            "Any X WHERE X firstname 'lulu' AND X name 'toto' OR X name 'tutu';"
        )
        base = tree.children[0].where
        self.assertEqual(isinstance(base, nodes.Or), 1)
        self.assertEqual(isinstance(base.children[0], nodes.And), 1)
        self.assertEqual(isinstance(base.children[1], nodes.Relation), 1)
        self.assertEqual(
            str(tree),
            'Any X WHERE (X firstname "lulu", X name "toto") OR (X name "tutu")',
        )

    def test_precedence_2(self):
        tree = self.parse(
            "Any X WHERE X firstname 'lulu', X name 'toto' OR X name 'tutu';"
        )
        base = tree.children[0].where
        self.assertEqual(isinstance(base, nodes.And), 1)
        self.assertEqual(isinstance(base.children[0], nodes.Relation), 1)
        self.assertEqual(isinstance(base.children[1], nodes.Or), 1)
        self.assertEqual(
            str(tree),
            'Any X WHERE X firstname "lulu", (X name "toto") OR (X name "tutu")',
        )

    def test_precedence_3(self):
        tree = self.parse(
            "Any X WHERE X firstname 'lulu' AND (X name 'toto' or X name 'tutu');"
        )
        base = tree.children[0].where
        self.assertEqual(isinstance(base, nodes.And), 1)
        self.assertEqual(isinstance(base.children[0], nodes.Relation), 1)
        self.assertEqual(isinstance(base.children[1], nodes.Or), 1)
        self.assertEqual(
            str(tree),
            'Any X WHERE X firstname "lulu", (X name "toto") OR (X name "tutu")',
        )

    def test_precedence_4(self):
        tree = self.parse(
            "Any X WHERE X firstname 'lulu' OR X name 'toto' AND X name 'tutu';"
        )
        base = tree.children[0].where
        self.assertEqual(isinstance(base, nodes.Or), 1)
        self.assertEqual(isinstance(base.children[0], nodes.Relation), 1)
        self.assertEqual(isinstance(base.children[1], nodes.And), 1)

    def test_not_precedence_0(self):
        tree = self.parse("Any X WHERE NOT X firstname 'lulu', X name 'toto';")
        self.assertEqual(str(tree), 'Any X WHERE NOT X firstname "lulu", X name "toto"')

    def test_not_precedence_1(self):
        tree = self.parse("Any X WHERE NOT X firstname 'lulu' AND X name 'toto';")
        self.assertEqual(str(tree), 'Any X WHERE NOT X firstname "lulu", X name "toto"')

    def test_not_precedence_2(self):
        tree = self.parse("Any X WHERE NOT X firstname 'lulu' OR X name 'toto';")
        self.assertEqual(
            str(tree), 'Any X WHERE (NOT X firstname "lulu") OR (X name "toto")'
        )

    def test_string_1(self):
        tree = self.parse(r"Any X WHERE X firstname 'lu\"lu';")
        const = tree.children[0].where.children[1].children[0]
        self.assertEqual(const.value, r"lu\"lu")

    def test_string_2(self):
        tree = self.parse(r"Any X WHERE X firstname 'lu\'lu';")
        const = tree.children[0].where.children[1].children[0]
        self.assertEqual(const.value, "lu'lu")

    def test_string_3(self):
        tree = self.parse(r'Any X WHERE X firstname "lu\'lu";')
        const = tree.children[0].where.children[1].children[0]
        self.assertEqual(const.value, r"lu\'lu")

    def test_string_4(self):
        tree = self.parse(r'Any X WHERE X firstname "lu\"lu";')
        const = tree.children[0].where.children[1].children[0]
        self.assertEqual(const.value, 'lu"lu')

    def test_math_1(self):
        tree = self.parse(r"Any X WHERE X date (TODAY + 1);")
        math = tree.children[0].where.children[1].children[0]
        self.assertIsInstance(math, nodes.MathExpression)
        self.assertEqual(math.operator, "+")

    def test_math_2(self):
        tree = self.parse(r"Any X WHERE X date (TODAY + 1 * 2);")
        math = tree.children[0].where.children[1].children[0]
        self.assertIsInstance(math, nodes.MathExpression)
        self.assertEqual(math.operator, "+")
        math2 = math.children[1]
        self.assertIsInstance(math2, nodes.MathExpression)
        self.assertEqual(math2.operator, "*")

    def test_math_3(self):
        tree = self.parse(r"Any X WHERE X date (TODAY + 1) * 2;")
        math = tree.children[0].where.children[1].children[0]
        self.assertIsInstance(math, nodes.MathExpression)
        self.assertEqual(math.operator, "*")
        math2 = math.children[0]
        self.assertIsInstance(math2, nodes.MathExpression)
        self.assertEqual(math2.operator, "+")

    def test_substitute(self):
        tree = self.parse("Any X WHERE X firstname %(firstname)s;")
        cste = tree.children[0].where.children[1].children[0]
        self.assertIsInstance(cste, nodes.Constant)
        self.assertEqual(cste.type, "Substitute")
        self.assertEqual(cste.value, "firstname")

    def test_optional_relation(self):
        tree = self.parse(r"Any X WHERE X related Y;")
        related = tree.children[0].where
        self.assertEqual(related.optional, None)
        tree = self.parse(r"Any X WHERE X? related Y;")
        related = tree.children[0].where
        self.assertEqual(related.optional, "left")
        tree = self.parse(r"Any X WHERE X related Y?;")
        related = tree.children[0].where
        self.assertEqual(related.optional, "right")
        tree = self.parse(r"Any X WHERE X? related Y?;")
        related = tree.children[0].where
        self.assertEqual(related.optional, "both")

    def test_exists(self):
        tree = self.parse(
            "Any X WHERE X firstname 'lulu',"
            "EXISTS (X owned_by U, U in_group G, G name 'lulufanclub' OR G name 'managers');"
        )
        self.assertEqual(
            tree.as_string(),
            'Any X WHERE X firstname "lulu", '
            'EXISTS(X owned_by U, U in_group G, (G name "lulufanclub") OR (G name "managers"))',
        )
        exists = tree.children[0].where.get_nodes(nodes.Exists)[0]
        self.assertIs(exists.children[0].parent, exists)
        self.assertTrue(exists.parent)

    def test_etype(self):
        tree = self.parse("EmailAddress X;")
        self.assertEqual(tree.as_string(), "Any X WHERE X is EmailAddress")
        tree = self.parse("EUser X;")
        self.assertEqual(tree.as_string(), "Any X WHERE X is EUser")

    def test_projection_with_attribute_to_get(self):
        tree = self.parse("Any X.my_attribute Where X is CWUser;")
        varref = tree.children[0].selection[0]
        self.assertIsInstance(varref, nodes.VariableRefAttributeAccess)
        self.assertEqual(varref.attribute_to_access, "my_attribute")

    def test_projection_with_function_to_apply(self):
        tree = self.parse('Any X.my_function("loutre") Where X is CWUser;')
        varref = tree.children[0].selection[0]
        self.assertIsInstance(varref, nodes.VariableRefMethodCall)
        self.assertEqual(varref.method_to_call, "my_function")
        self.assertEqual(varref.args_for_method_to_call[0].value, "loutre")

    def test_spec(self):
        """test all RQL string found in the specification and test they are well parsed"""
        for rql in SPEC_QUERIES:
            with self.subTest(rql=rql):
                self.parse(rql, True)

    def test_raise_badsyntax_error(self):
        for rql in BAD_SYNTAX_QUERIES:
            with self.subTest(rql=rql):
                self.assertRaises(self._syntaxerr, self.parse, rql)

    def test_raise_badrqlquery(self):
        BAD_QUERIES = ("Person Marcou;",)
        for rql in BAD_QUERIES:
            with self.subTest(rql=rql):
                self.assertRaises(BadRQLQuery, self.parse, rql)


class ParserRQLHelper(ParserHercule):
    _syntaxerr = RQLSyntaxError

    def parse(self, string, print_errors=False):
        try:
            return parse(string, print_errors)
        except Exception:
            raise


if __name__ == "__main__":
    unittest.main()



# pylint: disable-msg=W0622
# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.
"""RQL packaging information."""

import sys
import subprocess
import os.path as osp
from distutils.core import Extension

__docformat__ = "restructuredtext en"

modname = "rql"
numversion = (0, 42, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"

author = "Logilab"
author_email = "contact@logilab.fr"

description = "relationship query language (RQL) utilities"
long_desc = """A library providing the base utilities to handle RQL queries,
such as a parser, a type inferencer.
"""
web = "https://forge.extranet.logilab.fr/CubicWeb/RQL/"

classifiers = [
    "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Database",
    "Topic :: Software Development :: Libraries :: Python Modules",
]


def gecode_version():
    version = [6, 2, 0]
    if osp.exists("data/gecode_version.cc"):
        try:
            subprocess.check_call(
                ["g++", "-o", "gecode_version", "data/gecode_version.cc"]
            )
            p = subprocess.Popen("./gecode_version", stdout=subprocess.PIPE)
            vers = p.communicate()[0].decode("ascii")
            version = [int(c) for c in vers.strip().split(".")]
        except (EnvironmentError, subprocess.CalledProcessError):
            pass
    return version


def encode_version(a, b, c):
    return (a << 16) + (b << 8) + c


GECODE_VERSION = encode_version(*gecode_version())

if sys.platform != "win32":
    ext_modules = [
        Extension(
            "rql.rql_solve",
            ["rql/gecode_solver.cpp"],
            libraries=[
                "gecodeint",
                "gecodekernel",
                "gecodesearch",
            ],
            extra_compile_args=[f"-DGE_VERSION={GECODE_VERSION}"],
        )
    ]
else:
    ext_modules = [
        Extension(
            "rql.rql_solve",
            ["rql/gecode_solver.cpp"],
            libraries=[
                "GecodeInt-3-3-1-r-x86",
                "GecodeKernel-3-3-1-r-x86",
                "GecodeSearch-3-3-1-r-x86",
                "GecodeSupport-3-3-1-r-x86",
            ],
            extra_compile_args=[f"/DGE_VERSION={GECODE_VERSION}", "/EHsc"],
            # extra_link_args=['-static-libgcc'],
        )
    ]

install_requires = [
    "logilab-common >= 1.6.0",
    "logilab-database >= 1.6.0",
    "yapps2-logilab >= 2.2.1",
    "logilab-constraint >= 0.5.0",  # fallback if the gecode compiled module is missing
    'typing-extensions; python_version < "3.8"',
    "setuptools",
    'importlib_metadata; python_version < "3.10"',
]
